const sext = @import("zba-util").sext;

pub fn branch(comptime InstrFn: type, comptime L: bool) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u32) void {
            const cond: u4 = @truncate(opcode >> 28);
            switch (cond) {
                0b1111 => { // BLX
                    const H = L;
                    const offset = sext(u32, u24, opcode) << 2 | @as(u32, @intFromBool(H)) << 1;

                    cpu.r[14] = cpu.r[15] - 4;
                    cpu.cpsr.t.set();

                    cpu.r[15] +%= offset;
                },
                else => {
                    if (L) cpu.r[14] = cpu.r[15] - 4;

                    cpu.r[15] +%= sext(u32, u24, opcode) << 2;
                },
            }

            cpu.pipe.reload(cpu);
        }
    }.inner;
}

pub fn branchAndExchange(comptime InstrFn: type) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        pub fn inner(cpu: *Arm32, opcode: u32) void {
            const rn = opcode & 0xF;

            const thumb = cpu.r[rn] & 1 == 1;
            cpu.r[15] = cpu.r[rn] & if (thumb) ~@as(u32, 1) else ~@as(u32, 3);

            cpu.cpsr.t.write(thumb);
            cpu.pipe.reload(cpu);
        }
    }.inner;
}
